@@ -100,7 +100,7 @@ static const struct drive_list_entry drive_whitelist [] = {
 	{ "CONNER CTMA 4000"	,       "ALL"		},
 	{ "CONNER CTT8000-A"	,       "ALL"		},
 	{ "ST34342A"		,	"ALL"		},
-	{ 0			,	0		}
+	{ NULL			,	NULL		}
 };
 
 static const struct drive_list_entry drive_blacklist [] = {
@@ -125,7 +125,6 @@ static const struct drive_list_entry drive_blacklist [] = {
 	{ "HITACHI CDR-8335"	,	"ALL"		},
 	{ "HITACHI CDR-8435"	,	"ALL"		},
 	{ "Toshiba CD-ROM XM-6202B"	,	"ALL"		},
-	{ "TOSHIBA CD-ROM XM-1702BC"	,	"ALL"		},
 	{ "CD-532E-A"		,	"ALL"		},
 	{ "E-IDE CD-ROM CR-840",	"ALL"		},
 	{ "CD-ROM Drive/F5A",	"ALL"		},
@@ -139,7 +138,7 @@ static const struct drive_list_entry drive_blacklist [] = {
 	{ "PLEXTOR CD-R PX-W8432T",	"ALL"		},
 	{ "ATAPI CD-ROM DRIVE 40X MAXIMUM",	"ALL"		},
 	{ "_NEC DV5800A",               "ALL"           },  
-	{ 0			,	0		}
+	{ NULL			,	NULL		}
 
 };
 
@@ -514,9 +513,7 @@ int __ide_dma_off_quietly (ide_drive_t *drive)
 
 	if (HWIF(drive)->ide_dma_host_off(drive))
 		return 1;
-#ifdef CONFIG_BLK_DEV_IDE_TCQ
-	__ide_dma_queued_off(drive);
-#endif
+
 	return 0;
 }
 
