@@ -19,7 +19,7 @@
    TBD:
    * look at deferring rx frames rather than discarding (as per tulip)
    * handle tx ring full as per tulip
-   * performance test to tune rx_copybreak
+   * performace test to tune rx_copybreak
 
    Most of my modifications relate to the braindead big-endian
    implementation by Intel.  When the i596 is operating in
@@ -27,34 +27,26 @@
    should be stored as 0x56781234.  This is a real pain, when
    you have linked lists which are shared by the 680x0 and the
    i596.
-   NOTE: This was fixed in Stepping C of the chip. That again
-   means that big-endian should not be used at all: Steppings
-   A1, B and C of the 82596 each have a different big-endian behaviour.
 
    Driver skeleton
    Written 1993 by Donald Becker.
    Copyright 1993 United States Government as represented by the Director,
    National Security Agency. This software may only be used and distributed
-   according to the terms of the GNU Public License as modified by SRC,
+   according to the terms of the GNU General Public License as modified by SRC,
    incorporated herein by reference.
 
-   The author may be reached as becker@super.org or
-   C/O Supercomputing Research Ctr., 17100 Science Dr., Bowie MD 20715
+   The author may be reached as becker@scyld.com, or C/O
+   Scyld Computing Corporation, 410 Severn Ave., Suite 210, Annapolis MD 21403
 
  */
 
-static const char *version = "82596.c $Revision: 1.2 $\n";
-
 #include <linux/config.h>
 #include <linux/module.h>
-
 #include <linux/kernel.h>
-#include <linux/sched.h>
 #include <linux/string.h>
-#include <linux/ptrace.h>
 #include <linux/errno.h>
 #include <linux/ioport.h>
-#include <linux/malloc.h>
+#include <linux/slab.h>
 #include <linux/interrupt.h>
 #include <linux/delay.h>
 #include <linux/netdevice.h>
@@ -67,6 +59,11 @@ static const char *version = "82596.c $Revision: 1.2 $\n";
 #include <asm/dma.h>
 #include <asm/pgtable.h>
 
+static char version[] __initdata =
+	"82596.c $Revision: 1.5 $\n";
+
+#define DRV_NAME	"82596"
+
 /* DEBUG flags
  */
 
@@ -151,7 +148,10 @@ static int i596_debug = (DEB_SERIOUS|DEB_PROBE);
 
 MODULE_AUTHOR("Richard Hirst");
 MODULE_DESCRIPTION("i82596 driver");
+MODULE_LICENSE("GPL");
+
 MODULE_PARM(i596_debug, "i");
+MODULE_PARM_DESC(i596_debug, "i82596 debug mask");
 
 
 /* Copy frames shorter than rx_copybreak, otherwise pass on up in
@@ -191,6 +191,9 @@ enum commands {
 #define	 RX_SUSPEND	0x0030
 #define	 RX_ABORT	0x0040
 
+#define TX_TIMEOUT	5
+
+
 struct i596_reg {
 	unsigned short porthi;
 	unsigned short portlo;
@@ -329,9 +332,10 @@ struct i596_private {
 	struct tx_cmd tx_cmds[TX_RING_SIZE];
 	struct i596_tbd tbds[TX_RING_SIZE];
 	int next_tx_cmd;
+	spinlock_t lock;
 };
 
-char init_setup[] =
+static char init_setup[] =
 {
 	0x8E,			/* length, prefetch on */
 	0xC8,			/* fifo to 8, monitor off */
@@ -352,21 +356,22 @@ char init_setup[] =
 	0x00,
 	0x7f /*  *multi IA */ };
 
-static int i596_open(struct device *dev);
-static int i596_start_xmit(struct sk_buff *skb, struct device *dev);
-static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs);
-static int i596_close(struct device *dev);
-static struct net_device_stats *i596_get_stats(struct device *dev);
-static void i596_add_cmd(struct device *dev, struct i596_cmd *cmd);
+static int i596_open(struct net_device *dev);
+static int i596_start_xmit(struct sk_buff *skb, struct net_device *dev);
+static irqreturn_t i596_interrupt(int irq, void *dev_id, struct pt_regs *regs);
+static int i596_close(struct net_device *dev);
+static struct net_device_stats *i596_get_stats(struct net_device *dev);
+static void i596_add_cmd(struct net_device *dev, struct i596_cmd *cmd);
+static void i596_tx_timeout (struct net_device *dev);
 static void print_eth(unsigned char *buf, char *str);
-static void set_multicast_list(struct device *dev);
+static void set_multicast_list(struct net_device *dev);
 
 static int rx_ring_size = RX_RING_SIZE;
 static int ticks_limit = 25;
 static int max_cmd_backlog = TX_RING_SIZE-1;
-
 
-static inline void CA(struct device *dev)
+
+static inline void CA(struct net_device *dev)
 {
 #ifdef ENABLE_MVME16x_NET
 	if (MACH_IS_MVME16x) {
@@ -388,7 +393,7 @@ static inline void CA(struct device *dev)
 }
 
 
-static inline void MPU_PORT(struct device *dev, int c, volatile void *x)
+static inline void MPU_PORT(struct net_device *dev, int c, volatile void *x)
 {
 #ifdef ENABLE_MVME16x_NET
 	if (MACH_IS_MVME16x) {
@@ -409,12 +414,12 @@ static inline void MPU_PORT(struct device *dev, int c, volatile void *x)
 }
 
 
-static inline int wait_istat(struct device *dev, struct i596_private *lp, int delcnt, char *str)
+static inline int wait_istat(struct net_device *dev, struct i596_private *lp, int delcnt, char *str)
 {
 	while (--delcnt && lp->iscp.stat)
 		udelay(10);
 	if (!delcnt) {
-		printk("%s: %s, status %4.4x, cmd %4.4x.\n",
+		printk(KERN_ERR "%s: %s, status %4.4x, cmd %4.4x.\n",
 		     dev->name, str, lp->scb.status, lp->scb.command);
 		return -1;
 	}
@@ -423,12 +428,12 @@ static inline int wait_istat(struct device *dev, struct i596_private *lp, int de
 }
 
 
-static inline int wait_cmd(struct device *dev, struct i596_private *lp, int delcnt, char *str)
+static inline int wait_cmd(struct net_device *dev, struct i596_private *lp, int delcnt, char *str)
 {
 	while (--delcnt && lp->scb.command)
 		udelay(10);
 	if (!delcnt) {
-		printk("%s: %s, status %4.4x, cmd %4.4x.\n",
+		printk(KERN_ERR "%s: %s, status %4.4x, cmd %4.4x.\n",
 		     dev->name, str, lp->scb.status, lp->scb.command);
 		return -1;
 	}
@@ -437,44 +442,59 @@ static inline int wait_cmd(struct device *dev, struct i596_private *lp, int delc
 }
 
 
-static void i596_display_data(struct device *dev)
+static inline int wait_cfg(struct net_device *dev, struct i596_cmd *cmd, int delcnt, char *str)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	volatile struct i596_cmd *c = cmd;
+	
+	while (--delcnt && c->command)
+		udelay(10);
+	if (!delcnt) {
+		printk(KERN_ERR "%s: %s.\n", dev->name, str);
+		return -1;
+	}
+	else
+		return 0;
+}
+
+ 
+static void i596_display_data(struct net_device *dev)
+{
+	struct i596_private *lp = dev->priv;
 	struct i596_cmd *cmd;
 	struct i596_rfd *rfd;
 	struct i596_rbd *rbd;
 
-	printk("lp and scp at %p, .sysbus = %08lx, .iscp = %p\n",
+	printk(KERN_ERR "lp and scp at %p, .sysbus = %08lx, .iscp = %p\n",
 	       &lp->scp, lp->scp.sysbus, lp->scp.iscp);
-	printk("iscp at %p, iscp.stat = %08lx, .scb = %p\n",
+	printk(KERN_ERR "iscp at %p, iscp.stat = %08lx, .scb = %p\n",
 	       &lp->iscp, lp->iscp.stat, lp->iscp.scb);
-	printk("scb at %p, scb.status = %04x, .command = %04x,"
+	printk(KERN_ERR "scb at %p, scb.status = %04x, .command = %04x,"
 		" .cmd = %p, .rfd = %p\n",
 	       &lp->scb, lp->scb.status, lp->scb.command,
 		lp->scb.cmd, lp->scb.rfd);
-	printk("   errors: crc %lx, align %lx, resource %lx,"
+	printk(KERN_ERR "   errors: crc %lx, align %lx, resource %lx,"
                " over %lx, rcvdt %lx, short %lx\n",
 		lp->scb.crc_err, lp->scb.align_err, lp->scb.resource_err,
 		lp->scb.over_err, lp->scb.rcvdt_err, lp->scb.short_err);
 	cmd = lp->cmd_head;
 	while (cmd != I596_NULL) {
-		printk("cmd at %p, .status = %04x, .command = %04x, .b_next = %p\n",
+		printk(KERN_ERR "cmd at %p, .status = %04x, .command = %04x, .b_next = %p\n",
 		  cmd, cmd->status, cmd->command, cmd->b_next);
 		cmd = cmd->v_next;
 	}
 	rfd = lp->rfd_head;
-	printk("rfd_head = %p\n", rfd);
+	printk(KERN_ERR "rfd_head = %p\n", rfd);
 	do {
-		printk ("   %p .stat %04x, .cmd %04x, b_next %p, rbd %p,"
+		printk(KERN_ERR "   %p .stat %04x, .cmd %04x, b_next %p, rbd %p,"
                         " count %04x\n",
 			rfd, rfd->stat, rfd->cmd, rfd->b_next, rfd->rbd,
 			rfd->count);
 		rfd = rfd->v_next;
 	} while (rfd != lp->rfd_head);
 	rbd = lp->rbd_head;
-	printk("rbd_head = %p\n", rbd);
+	printk(KERN_ERR "rbd_head = %p\n", rbd);
 	do {
-		printk("   %p .count %04x, b_next %p, b_data %p, size %04x\n",
+		printk(KERN_ERR "   %p .count %04x, b_next %p, b_data %p, size %04x\n",
 			rbd, rbd->count, rbd->b_next, rbd->b_data, rbd->size);
 		rbd = rbd->v_next;
 	} while (rbd != lp->rbd_head);
@@ -482,21 +502,34 @@ static void i596_display_data(struct device *dev)
 
 
 #if defined(ENABLE_MVME16x_NET) || defined(ENABLE_BVME6000_NET)
-static void i596_error(int irq, void *dev_id, struct pt_regs *regs)
+static irqreturn_t i596_error(int irq, void *dev_id, struct pt_regs *regs)
 {
-	struct device *dev = dev_id;
-	volatile unsigned char *pcc2 = (unsigned char *) 0xfff42000;
+	struct net_device *dev = dev_id;
+#ifdef ENABLE_MVME16x_NET
+	if (MACH_IS_MVME16x) {
+		volatile unsigned char *pcc2 = (unsigned char *) 0xfff42000;
 
-	pcc2[0x28] = 1;
-	pcc2[0x2b] = 0x1d;
-	printk("%s: Error interrupt\n", dev->name);
+		pcc2[0x28] = 1;
+		pcc2[0x2b] = 0x1d;
+	}
+#endif
+#ifdef ENABLE_BVME6000_NET
+	if (MACH_IS_BVME6000) {
+		volatile unsigned char *ethirq = (unsigned char *) BVME_ETHIRQ_REG;
+
+		*ethirq = 1;
+		*ethirq = 3;
+	}
+#endif
+	printk(KERN_ERR "%s: Error interrupt\n", dev->name);
 	i596_display_data(dev);
+	return IRQ_HANDLED;
 }
 #endif
 
-static inline void init_rx_bufs(struct device *dev)
+static inline void init_rx_bufs(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *)dev->priv;
+	struct i596_private *lp = dev->priv;
 	int i;
 	struct i596_rfd *rfd;
 	struct i596_rbd *rbd;
@@ -545,9 +578,9 @@ static inline void init_rx_bufs(struct device *dev)
 	rfd->cmd = CMD_EOL|CMD_FLEX;
 }
 
-static inline void remove_rx_bufs(struct device *dev)
+static inline void remove_rx_bufs(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *)dev->priv;
+	struct i596_private *lp = dev->priv;
 	struct i596_rbd *rbd;
 	int i;
 
@@ -559,9 +592,9 @@ static inline void remove_rx_bufs(struct device *dev)
 }
 
 
-static void rebuild_rx_bufs(struct device *dev)
+static void rebuild_rx_bufs(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 	int i;
 
 	/* Ensure rx frame/buffer descriptors are tidy */
@@ -578,15 +611,15 @@ static void rebuild_rx_bufs(struct device *dev)
 }
 
 
-static int init_i596_mem(struct device *dev)
+static int init_i596_mem(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 #if !defined(ENABLE_MVME16x_NET) && !defined(ENABLE_BVME6000_NET)
 	short ioaddr = dev->base_addr;
 #endif
 	unsigned long flags;
 
-	MPU_PORT(dev, PORT_RESET, 0);
+	MPU_PORT(dev, PORT_RESET, NULL);
 
 	udelay(100);		/* Wait 100us - seems to help */
 
@@ -615,7 +648,7 @@ static int init_i596_mem(struct device *dev)
 
 	/* change the scp address */
 
-	MPU_PORT(dev, PORT_ALTSCP, (void *)virt_to_bus(&lp->scp));
+	MPU_PORT(dev, PORT_ALTSCP, (void *)virt_to_bus((void *)&lp->scp));
 
 #elif defined(ENABLE_APRICOT)
 
@@ -646,14 +679,21 @@ static int init_i596_mem(struct device *dev)
 		lp->scp.sysbus = 0x00440000;
 #endif
 
-	lp->scp.iscp = WSWAPiscp(virt_to_bus(&(lp->iscp)));
-	lp->iscp.scb = WSWAPscb(virt_to_bus(&(lp->scb)));
+	lp->scp.iscp = WSWAPiscp(virt_to_bus((void *)&lp->iscp));
+	lp->iscp.scb = WSWAPscb(virt_to_bus((void *)&lp->scb));
 	lp->iscp.stat = ISCP_BUSY;
 	lp->cmd_backlog = 0;
 
 	lp->cmd_head = lp->scb.cmd = I596_NULL;
 
-	DEB(DEB_INIT,printk("%s: starting i82596.\n", dev->name));
+#ifdef ENABLE_BVME6000_NET
+	if (MACH_IS_BVME6000) {
+		lp->scb.t_on  = 7 * 25;
+		lp->scb.t_off = 1 * 25;
+	}
+#endif
+
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: starting i82596.\n", dev->name));
 
 #if defined(ENABLE_APRICOT)
 	(void) inb(ioaddr + 0x10);
@@ -663,7 +703,7 @@ static int init_i596_mem(struct device *dev)
 
 	if (wait_istat(dev,lp,1000,"initialization timed out"))
 		goto failed;
-	DEB(DEB_INIT,printk("%s: i82596 initialization successful\n", dev->name));
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: i82596 initialization successful\n", dev->name));
 
 	/* Ensure rx frame/buffer descriptors are tidy */
 	rebuild_rx_bufs(dev);
@@ -687,51 +727,51 @@ static int init_i596_mem(struct device *dev)
 #endif
 
 
-	DEB(DEB_INIT,printk("%s: queuing CmdConfigure\n", dev->name));
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: queuing CmdConfigure\n", dev->name));
 	memcpy(lp->cf_cmd.i596_config, init_setup, 14);
 	lp->cf_cmd.cmd.command = CmdConfigure;
 	i596_add_cmd(dev, &lp->cf_cmd.cmd);
 
-	DEB(DEB_INIT,printk("%s: queuing CmdSASetup\n", dev->name));
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: queuing CmdSASetup\n", dev->name));
 	memcpy(lp->sa_cmd.eth_addr, dev->dev_addr, 6);
 	lp->sa_cmd.cmd.command = CmdSASetup;
 	i596_add_cmd(dev, &lp->sa_cmd.cmd);
 
-	DEB(DEB_INIT,printk("%s: queuing CmdTDR\n", dev->name));
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: queuing CmdTDR\n", dev->name));
 	lp->tdr_cmd.cmd.command = CmdTDR;
 	i596_add_cmd(dev, &lp->tdr_cmd.cmd);
 
+	spin_lock_irqsave (&lp->lock, flags);
 
-	save_flags(flags);
-	cli();
-
-	if (wait_cmd(dev,lp,1000,"timed out waiting to issue RX_START"))
+	if (wait_cmd(dev,lp,1000,"timed out waiting to issue RX_START")) {
+		spin_unlock_irqrestore (&lp->lock, flags);
 		goto failed;
-	DEB(DEB_INIT,printk("%s: Issuing RX_START\n", dev->name));
+	}
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: Issuing RX_START\n", dev->name));
 	lp->scb.command = RX_START;
 	CA(dev);
 
-	restore_flags(flags);
+	spin_unlock_irqrestore (&lp->lock, flags);
 
 	if (wait_cmd(dev,lp,1000,"RX_START not processed"))
 		goto failed;
-	DEB(DEB_INIT,printk("%s: Receive unit started OK\n", dev->name));
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: Receive unit started OK\n", dev->name));
 	return 0;
 
 failed:
-	printk("%s: Failed to initialise 82596\n", dev->name);
-	MPU_PORT(dev, PORT_RESET, 0);
+	printk(KERN_CRIT "%s: Failed to initialise 82596\n", dev->name);
+	MPU_PORT(dev, PORT_RESET, NULL);
 	return -1;
 }
 
-static inline int i596_rx(struct device *dev)
+static inline int i596_rx(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *)dev->priv;
+	struct i596_private *lp = dev->priv;
 	struct i596_rfd *rfd;
 	struct i596_rbd *rbd;
 	int frames = 0;
 
-	DEB(DEB_RXFRAME,printk ("i596_rx(), rfd_head %p, rbd_head %p\n",
+	DEB(DEB_RXFRAME,printk(KERN_DEBUG "i596_rx(), rfd_head %p, rbd_head %p\n",
 			lp->rfd_head, lp->rbd_head));
 
 	rfd = lp->rfd_head;		/* Ref next frame to check */
@@ -742,11 +782,11 @@ static inline int i596_rx(struct device *dev)
 		else if (rfd->rbd == lp->rbd_head->b_addr)
 			rbd = lp->rbd_head;
 		else {
-			printk("%s: rbd chain broken!\n", dev->name);
+			printk(KERN_CRIT "%s: rbd chain broken!\n", dev->name);
 			/* XXX Now what? */
 			rbd = I596_NULL;
 		}
-		DEB(DEB_RXFRAME, printk("  rfd %p, rfd.rbd %p, rfd.stat %04x\n",
+		DEB(DEB_RXFRAME, printk(KERN_DEBUG "  rfd %p, rfd.rbd %p, rfd.stat %04x\n",
 			rfd, rfd->rbd, rfd->stat));
 		
 		if (rbd != I596_NULL && ((rfd->stat) & STAT_OK)) {
@@ -787,7 +827,7 @@ static inline int i596_rx(struct device *dev)
 memory_squeeze:
 			if (skb == NULL) {
 				/* XXX tulip.c can defer packets here!! */
-				printk ("%s: i596_rx Memory squeeze, dropping packet.\n", dev->name);
+				printk(KERN_WARNING "%s: i596_rx Memory squeeze, dropping packet.\n", dev->name);
 				lp->stats.rx_dropped++;
 			}
 			else {
@@ -804,12 +844,13 @@ memory_squeeze:
 						pkt_len);
 #endif
 				netif_rx(skb);
+				dev->last_rx = jiffies;
 				lp->stats.rx_packets++;
 				lp->stats.rx_bytes+=pkt_len;
 			}
 		}
 		else {
-			DEB(DEB_ERRORS, printk("%s: Error, rfd.stat = 0x%04x\n",
+			DEB(DEB_ERRORS, printk(KERN_DEBUG "%s: Error, rfd.stat = 0x%04x\n",
 					dev->name, rfd->stat));
 			lp->stats.rx_errors++;
 			if ((rfd->stat) & 0x0001)
@@ -853,13 +894,13 @@ memory_squeeze:
 		rfd = lp->rfd_head;
 	}
 
-	DEB(DEB_RXFRAME,printk ("frames %d\n", frames));
+	DEB(DEB_RXFRAME,printk(KERN_DEBUG "frames %d\n", frames));
 
 	return 0;
 }
 
 
-static inline void i596_cleanup_cmd(struct device *dev, struct i596_private *lp)
+static inline void i596_cleanup_cmd(struct net_device *dev, struct i596_private *lp)
 {
 	struct i596_cmd *ptr;
 
@@ -892,58 +933,45 @@ static inline void i596_cleanup_cmd(struct device *dev, struct i596_private *lp)
 	lp->scb.cmd = I596_NULL;
 }
 
-static inline void i596_reset(struct device *dev, struct i596_private *lp, int ioaddr)
+static inline void i596_reset(struct net_device *dev, struct i596_private *lp, int ioaddr)
 {
 	unsigned long flags;
 
-	DEB(DEB_RESET,printk("i596_reset\n"));
+	DEB(DEB_RESET,printk(KERN_DEBUG "i596_reset\n"));
 
-	save_flags(flags);
-	cli();
+	spin_lock_irqsave (&lp->lock, flags);
 
 	wait_cmd(dev,lp,100,"i596_reset timed out");
-	dev->start = 0;
-	dev->tbusy = 1;
+
+	netif_stop_queue(dev);
 
 	lp->scb.command = CUC_ABORT | RX_ABORT;
 	CA(dev);
 
 	/* wait for shutdown */
 	wait_cmd(dev,lp,1000,"i596_reset 2 timed out");
-	restore_flags(flags);
+	spin_unlock_irqrestore (&lp->lock, flags);
 
 	i596_cleanup_cmd(dev,lp);
 	i596_rx(dev);
 
-	dev->start = 1;
-	dev->tbusy = 0;
-	dev->interrupt = 0;
+	netif_start_queue(dev);
 	init_i596_mem(dev);
 }
 
-static void i596_add_cmd(struct device *dev, struct i596_cmd *cmd)
+static void i596_add_cmd(struct net_device *dev, struct i596_cmd *cmd)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 	int ioaddr = dev->base_addr;
 	unsigned long flags;
 
-	DEB(DEB_ADDCMD,printk("i596_add_cmd\n"));
+	DEB(DEB_ADDCMD,printk(KERN_DEBUG "i596_add_cmd\n"));
 
 	cmd->status = 0;
 	cmd->command |= (CMD_EOL | CMD_INTR);
 	cmd->v_next = cmd->b_next = I596_NULL;
-	save_flags(flags);
-	cli();
 
-	/*
-	 * RGH  300597:  Looks to me like there could be a race condition
-	 * here.  Just because we havn't picked up all the command items
-	 * yet, doesn't mean that the 82596 hasn't finished processing
-	 * them.  So, we may need to do a CUC_START anyway.
-	 * Maybe not.  If it interrupts saying the CU is idle when there
-	 * is still something in the cmd queue, the int handler with restart
-	 * the CU.
-	 */
+	spin_lock_irqsave (&lp->lock, flags);
 
 	if (lp->cmd_head != I596_NULL) {
 		lp->cmd_tail->v_next = cmd;
@@ -958,7 +986,7 @@ static void i596_add_cmd(struct device *dev, struct i596_cmd *cmd)
 	lp->cmd_tail = cmd;
 	lp->cmd_backlog++;
 
-	restore_flags(flags);
+	spin_unlock_irqrestore (&lp->lock, flags);
 
 	if (lp->cmd_backlog > max_cmd_backlog) {
 		unsigned long tickssofar = jiffies - lp->last_cmd;
@@ -966,34 +994,31 @@ static void i596_add_cmd(struct device *dev, struct i596_cmd *cmd)
 		if (tickssofar < ticks_limit)
 			return;
 
-		printk("%s: command unit timed out, status resetting.\n", dev->name);
+		printk(KERN_NOTICE "%s: command unit timed out, status resetting.\n", dev->name);
 
 		i596_reset(dev, lp, ioaddr);
 	}
 }
 
-static int i596_open(struct device *dev)
+static int i596_open(struct net_device *dev)
 {
 	int res = 0;
 
-	DEB(DEB_OPEN,printk("%s: i596_open() irq %d.\n", dev->name, dev->irq));
+	DEB(DEB_OPEN,printk(KERN_DEBUG "%s: i596_open() irq %d.\n", dev->name, dev->irq));
 
-	if (request_irq(dev->irq, &i596_interrupt, 0, "i82596", dev)) {
-		printk("%s: IRQ %d not free\n", dev->name, dev->irq);
+	if (request_irq(dev->irq, i596_interrupt, 0, "i82596", dev)) {
+		printk(KERN_ERR "%s: IRQ %d not free\n", dev->name, dev->irq);
 		return -EAGAIN;
 	}
 #ifdef ENABLE_MVME16x_NET
 	if (MACH_IS_MVME16x) {
-		if (request_irq(0x56, &i596_error, 0, "i82596_error", dev))
+		if (request_irq(0x56, i596_error, 0, "i82596_error", dev))
 			return -EAGAIN;
 	}
 #endif
 	init_rx_bufs(dev);
 
-	dev->tbusy = 0;
-	dev->interrupt = 0;
-	dev->start = 1;
-	MOD_INC_USE_COUNT;
+	netif_start_queue(dev);
 
 	/* Initialize the 82596 memory */
 	if (init_i596_mem(dev)) {
@@ -1004,87 +1029,90 @@ static int i596_open(struct device *dev)
 	return res;
 }
 
-static int i596_start_xmit(struct sk_buff *skb, struct device *dev)
+static void i596_tx_timeout (struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 	int ioaddr = dev->base_addr;
-	struct tx_cmd *tx_cmd;
-	struct i596_tbd *tbd;
 
 	/* Transmitter timeout, serious problems. */
-	if (dev->tbusy) {
-		int tickssofar = jiffies - dev->trans_start;
-		if (tickssofar < 5)
-			return 1;
-		DEB(DEB_SERIOUS,printk("%s: transmit timed out, status resetting.\n",
-		       dev->name));
-		lp->stats.tx_errors++;
-		/* Try to restart the adaptor */
-		if (lp->last_restart == lp->stats.tx_packets) {
-			DEB(DEB_SERIOUS,printk("%s: Resetting board.\n",
-					dev->name));
+	DEB(DEB_ERRORS,printk(KERN_ERR "%s: transmit timed out, status resetting.\n",
+			dev->name));
 
-			/* Shutdown and restart */
-			i596_reset(dev, lp, ioaddr);
-		} else {
-			/* Issue a channel attention signal */
-			DEB(DEB_SERIOUS,printk("%s: Kicking board.\n",
-					dev->name));
-			lp->scb.command = CUC_START | RX_START;
-			CA(dev);
-			lp->last_restart = lp->stats.tx_packets;
-		}
-		dev->tbusy = 0;
-		dev->trans_start = jiffies;
+	lp->stats.tx_errors++;
+
+	/* Try to restart the adaptor */
+	if (lp->last_restart == lp->stats.tx_packets) {
+		DEB(DEB_ERRORS,printk(KERN_ERR "Resetting board.\n"));
+		/* Shutdown and restart */
+		i596_reset (dev, lp, ioaddr);
+	} else {
+		/* Issue a channel attention signal */
+		DEB(DEB_ERRORS,printk(KERN_ERR "Kicking board.\n"));
+		lp->scb.command = CUC_START | RX_START;
+		CA (dev);
+		lp->last_restart = lp->stats.tx_packets;
 	}
-	DEB(DEB_STARTTX,printk("%s: i596_start_xmit(%x,%x) called\n", dev->name,
+
+	dev->trans_start = jiffies;
+	netif_wake_queue (dev);
+}
+
+
+static int i596_start_xmit(struct sk_buff *skb, struct net_device *dev)
+{
+	struct i596_private *lp = dev->priv;
+	struct tx_cmd *tx_cmd;
+	struct i596_tbd *tbd;
+	short length = skb->len;
+	dev->trans_start = jiffies;
+
+	DEB(DEB_STARTTX,printk(KERN_DEBUG "%s: i596_start_xmit(%x,%x) called\n", dev->name,
 				skb->len, (unsigned int)skb->data));
 
-	/* Block a timer-based transmit from overlapping.  This could better be
-	   done with atomic_swap(1, dev->tbusy), but set_bit() works as well. */
-	if (test_and_set_bit(0, (void *) &dev->tbusy) != 0)
-		printk("%s: Transmitter access conflict.\n", dev->name);
-	else {
-		short length = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
-		dev->trans_start = jiffies;
+	if (skb->len < ETH_ZLEN) {
+		skb = skb_padto(skb, ETH_ZLEN);
+		if (skb == NULL)
+			return 0;
+		length = ETH_ZLEN;
+	}
+	netif_stop_queue(dev);
 
-		tx_cmd = lp->tx_cmds + lp->next_tx_cmd;
-		tbd = lp->tbds + lp->next_tx_cmd;
+	tx_cmd = lp->tx_cmds + lp->next_tx_cmd;
+	tbd = lp->tbds + lp->next_tx_cmd;
 
-		if (tx_cmd->cmd.command) {
-			DEB(DEB_ERRORS,printk ("%s: xmit ring full, dropping packet.\n",
-					dev->name));
-			lp->stats.tx_dropped++;
+	if (tx_cmd->cmd.command) {
+		printk(KERN_NOTICE "%s: xmit ring full, dropping packet.\n",
+				dev->name);
+		lp->stats.tx_dropped++;
 
-			dev_kfree_skb(skb);
-		} else {
-			if (++lp->next_tx_cmd == TX_RING_SIZE)
-				lp->next_tx_cmd = 0;
-			tx_cmd->tbd = WSWAPtbd(virt_to_bus(tbd));
-			tbd->next = I596_NULL;
+		dev_kfree_skb(skb);
+	} else {
+		if (++lp->next_tx_cmd == TX_RING_SIZE)
+			lp->next_tx_cmd = 0;
+		tx_cmd->tbd = WSWAPtbd(virt_to_bus(tbd));
+		tbd->next = I596_NULL;
 
-			tx_cmd->cmd.command = CMD_FLEX | CmdTx;
-			tx_cmd->skb = skb;
+		tx_cmd->cmd.command = CMD_FLEX | CmdTx;
+		tx_cmd->skb = skb;
 
-			tx_cmd->pad = 0;
-			tx_cmd->size = 0;
-			tbd->pad = 0;
-			tbd->size = EOF | length;
+		tx_cmd->pad = 0;
+		tx_cmd->size = 0;
+		tbd->pad = 0;
+		tbd->size = EOF | length;
 
-			tbd->data = WSWAPchar(virt_to_bus(skb->data));
+		tbd->data = WSWAPchar(virt_to_bus(skb->data));
 
 #ifdef __mc68000__
-			cache_push(virt_to_phys(skb->data), length);
+		cache_push(virt_to_phys(skb->data), length);
 #endif
-			DEB(DEB_TXADDR,print_eth(skb->data, "tx-queued"));
-			i596_add_cmd(dev, &tx_cmd->cmd);
+		DEB(DEB_TXADDR,print_eth(skb->data, "tx-queued"));
+		i596_add_cmd(dev, &tx_cmd->cmd);
 
-			lp->stats.tx_packets++;
-			lp->stats.tx_bytes += length;
-		}
+		lp->stats.tx_packets++;
+		lp->stats.tx_bytes += length;
 	}
 
-	dev->tbusy = 0;
+	netif_start_queue(dev);
 
 	return 0;
 }
@@ -1093,7 +1121,7 @@ static void print_eth(unsigned char *add, char *str)
 {
 	int i;
 
-	printk("i596 0x%p, ", add);
+	printk(KERN_DEBUG "i596 0x%p, ", add);
 	for (i = 0; i < 6; i++)
 		printk(" %02X", add[i + 6]);
 	printk(" -->");
@@ -1102,24 +1130,41 @@ static void print_eth(unsigned char *add, char *str)
 	printk(" %02X%02X, %s\n", add[12], add[13], str);
 }
 
-__initfunc(int i82596_probe(struct device *dev))
+static int io = 0x300;
+static int irq = 10;
+
+struct net_device * __init i82596_probe(int unit)
 {
+	struct net_device *dev;
 	int i;
 	struct i596_private *lp;
-	char eth_addr[6];
-	static int probed = 0;
+	char eth_addr[8];
+	static int probed;
+	int err;
 
 	if (probed)
-		return -ENODEV;
+		return ERR_PTR(-ENODEV);
 	probed++;
+
+	dev = alloc_etherdev(0);
+	if (!dev)
+		return ERR_PTR(-ENOMEM);
+
+	if (unit >= 0) {
+		sprintf(dev->name, "eth%d", unit);
+		netdev_boot_setup_check(dev);
+	} else {
+		dev->base_addr = io;
+		dev->irq = irq;
+	}
+
 #ifdef ENABLE_MVME16x_NET
 	if (MACH_IS_MVME16x) {
-#ifdef XXX_FIXME
 		if (mvme16x_config & MVME16x_CONFIG_NO_ETHERNET) {
-			printk("Ethernet probe disabled - chip not present\n");
-			return -ENODEV;
+			printk(KERN_NOTICE "Ethernet probe disabled - chip not present\n");
+			err = -ENODEV;
+			goto out;
 		}
-#endif
 		memcpy(eth_addr, (void *) 0xfffc1f2c, 6);	/* YUCK! Get addr from NOVRAM */
 		dev->base_addr = MVME_I596_BASE;
 		dev->irq = (unsigned) MVME16x_IRQ_I596;
@@ -1147,9 +1192,10 @@ __initfunc(int i82596_probe(struct device *dev))
 		/* this is easy the ethernet interface can only be at 0x300 */
 		/* first check nothing is already registered here */
 
-		if (check_region(ioaddr, I596_TOTAL_SIZE)) {
-			printk("82596: IO address 0x%04x in use\n", ioaddr);
-			return -ENODEV;
+		if (!request_region(ioaddr, I596_TOTAL_SIZE, DRV_NAME)) {
+			printk(KERN_ERR "82596: IO address 0x%04x in use\n", ioaddr);
+			err = -EBUSY;
+			goto out;
 		}
 
 		for (i = 0; i < 8; i++) {
@@ -1159,18 +1205,15 @@ __initfunc(int i82596_probe(struct device *dev))
 
 		/* checksum is a multiple of 0x100, got this wrong first time
 		   some machines have 0x100, some 0x200. The DOS driver doesn't
-		   even bother with the checksum */
+		   even bother with the checksum.
+		   Some other boards trip the checksum.. but then appear as
+		   ether address 0. Trap these - AC */
 
-		if (checksum % 0x100)
-			return -ENODEV;
-
-		/* Some other boards trip the checksum.. but then appear as
-		 * ether address 0. Trap these - AC */
-
-		if (memcmp(eth_addr, "\x00\x00\x49", 3) != 0)
-			return -ENODEV;
-
-		request_region(ioaddr, I596_TOTAL_SIZE, "i596");
+		if ((checksum % 0x100) || 
+		    (memcmp(eth_addr, "\x00\x00\x49", 3) != 0)) {
+			err = -ENODEV;
+			goto out1;
+		}
 
 		dev->base_addr = ioaddr;
 		dev->irq = 10;
@@ -1178,33 +1221,33 @@ __initfunc(int i82596_probe(struct device *dev))
 #endif
 	dev->mem_start = (int)__get_free_pages(GFP_ATOMIC, 0);
 	if (!dev->mem_start) {
-#ifdef ENABLE_APRICOT
-		release_region(dev->base_addr, I596_TOTAL_SIZE);
-#endif
-		return -ENOMEM;
+		err = -ENOMEM;
+		goto out1;
 	}
 
-	ether_setup(dev);
-	DEB(DEB_PROBE,printk("%s: 82596 at %#3lx,", dev->name, dev->base_addr));
+	DEB(DEB_PROBE,printk(KERN_INFO "%s: 82596 at %#3lx,", dev->name, dev->base_addr));
 
 	for (i = 0; i < 6; i++)
 		DEB(DEB_PROBE,printk(" %2.2X", dev->dev_addr[i] = eth_addr[i]));
 
 	DEB(DEB_PROBE,printk(" IRQ %d.\n", dev->irq));
 
-	DEB(DEB_PROBE,printk(version));
+	DEB(DEB_PROBE,printk(KERN_INFO "%s", version));
 
 	/* The 82596-specific entries in the device structure. */
-	dev->open = &i596_open;
-	dev->stop = &i596_close;
-	dev->hard_start_xmit = &i596_start_xmit;
-	dev->get_stats = &i596_get_stats;
-	dev->set_multicast_list = &set_multicast_list;
+	SET_MODULE_OWNER(dev);
+	dev->open = i596_open;
+	dev->stop = i596_close;
+	dev->hard_start_xmit = i596_start_xmit;
+	dev->get_stats = i596_get_stats;
+	dev->set_multicast_list = set_multicast_list;
+	dev->tx_timeout = i596_tx_timeout;
+	dev->watchdog_timeo = TX_TIMEOUT;
 
 	dev->priv = (void *)(dev->mem_start);
 
-	lp = (struct i596_private *) dev->priv;
-	DEB(DEB_INIT,printk ("%s: lp at 0x%08lx (%d bytes), lp->scb at 0x%08lx\n",
+	lp = dev->priv;
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: lp at 0x%08lx (%d bytes), lp->scb at 0x%08lx\n",
 			dev->name, (unsigned long)lp,
 			sizeof(struct i596_private), (unsigned long)&lp->scb));
 	memset((void *) lp, 0, sizeof(struct i596_private));
@@ -1217,43 +1260,60 @@ __initfunc(int i82596_probe(struct device *dev))
 	lp->scb.command = 0;
 	lp->scb.cmd = I596_NULL;
 	lp->scb.rfd = I596_NULL;
+	lp->lock = SPIN_LOCK_UNLOCKED;
 
-	return 0;
+	err = register_netdev(dev);
+	if (err)
+		goto out2;
+	return dev;
+out2:
+#ifdef __mc68000__
+	/* XXX This assumes default cache mode to be IOMAP_FULL_CACHING,
+	 * XXX which may be invalid (CONFIG_060_WRITETHROUGH)
+	 */
+	kernel_set_cachemode((void *)(dev->mem_start), 4096,
+			IOMAP_FULL_CACHING);
+#endif
+	free_page ((u32)(dev->mem_start));
+out1:
+#ifdef ENABLE_APRICOT
+	release_region(dev->base_addr, I596_TOTAL_SIZE);
+#endif
+out:
+	free_netdev(dev);
+	return ERR_PTR(err);
 }
 
-static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
+static irqreturn_t i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 {
-	struct device *dev = dev_id;
+	struct net_device *dev = dev_id;
 	struct i596_private *lp;
 	short ioaddr;
 	unsigned short status, ack_cmd = 0;
+	int handled = 0;
 
 #ifdef ENABLE_BVME6000_NET
 	if (MACH_IS_BVME6000) {
 		if (*(char *) BVME_LOCAL_IRQ_STAT & BVME_ETHERR) {
-			i596_error(BVME_IRQ_I596, NULL, NULL);
-			return;
+			i596_error(irq, dev_id, regs);
+			return IRQ_HANDLED;
 		}
 	}
 #endif
 	if (dev == NULL) {
-		printk("i596_interrupt(): irq %d for unknown device.\n", irq);
-		return;
+		printk(KERN_ERR "i596_interrupt(): irq %d for unknown device.\n", irq);
+		return IRQ_NONE;
 	}
 
-	if (dev->interrupt)
-		printk("%s: Re-entering the interrupt handler.\n", dev->name);
-
-	dev->interrupt = 1;
-
 	ioaddr = dev->base_addr;
+	lp = dev->priv;
 
-	lp = (struct i596_private *) dev->priv;
+	spin_lock (&lp->lock);
 
 	wait_cmd(dev,lp,100,"i596 interrupt, timeout");
 	status = lp->scb.status;
 
-	DEB(DEB_INTS,printk("%s: i596 interrupt, IRQ %d, status %4.4x.\n",
+	DEB(DEB_INTS,printk(KERN_DEBUG "%s: i596 interrupt, IRQ %d, status %4.4x.\n",
 			dev->name, irq, status));
 
 	ack_cmd = status & 0xf000;
@@ -1261,15 +1321,16 @@ static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 	if ((status & 0x8000) || (status & 0x2000)) {
 		struct i596_cmd *ptr;
 
+		handled = 1;
 		if ((status & 0x8000))
-			DEB(DEB_INTS,printk("%s: i596 interrupt completed command.\n", dev->name));
+			DEB(DEB_INTS,printk(KERN_DEBUG "%s: i596 interrupt completed command.\n", dev->name));
 		if ((status & 0x2000))
-			DEB(DEB_INTS,printk("%s: i596 interrupt command unit inactive %x.\n", dev->name, status & 0x0700));
+			DEB(DEB_INTS,printk(KERN_DEBUG "%s: i596 interrupt command unit inactive %x.\n", dev->name, status & 0x0700));
 
 		while ((lp->cmd_head != I596_NULL) && (lp->cmd_head->status & STAT_C)) {
 			ptr = lp->cmd_head;
 
-			DEB(DEB_STATUS,printk("cmd_head->status = %04x, ->command = %04x\n",
+			DEB(DEB_STATUS,printk(KERN_DEBUG "cmd_head->status = %04x, ->command = %04x\n",
 				       lp->cmd_head->status, lp->cmd_head->command));
 			lp->cmd_head = ptr->v_next;
 			lp->cmd_backlog--;
@@ -1296,7 +1357,7 @@ static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 						lp->stats.tx_aborted_errors++;
 				}
 
-				dev_kfree_skb(skb);
+				dev_kfree_skb_irq(skb);
 
 				tx_cmd->cmd.command = 0; /* Mark free */
 				break;
@@ -1306,20 +1367,21 @@ static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 				unsigned short status = ((struct tdr_cmd *)ptr)->status;
 
 				if (status & 0x8000) {
-					DEB(DEB_ANY,printk("%s: link ok.\n", dev->name));
+					DEB(DEB_TDR,printk(KERN_INFO "%s: link ok.\n", dev->name));
 				} else {
 					if (status & 0x4000)
-						printk("%s: Transceiver problem.\n", dev->name);
+						printk(KERN_ERR "%s: Transceiver problem.\n", dev->name);
 					if (status & 0x2000)
-						printk("%s: Termination problem.\n", dev->name);
+						printk(KERN_ERR "%s: Termination problem.\n", dev->name);
 					if (status & 0x1000)
-						printk("%s: Short circuit.\n", dev->name);
+						printk(KERN_ERR "%s: Short circuit.\n", dev->name);
 
-					DEB(DEB_TDR,printk("%s: Time %d.\n", dev->name, status & 0x07ff));
+					DEB(DEB_TDR,printk(KERN_INFO "%s: Time %d.\n", dev->name, status & 0x07ff));
 				}
 				break;
 			    }
 			case CmdConfigure:
+			case CmdMulticastList:
 				/* Zap command so set_multicast_list() knows it is free */
 				ptr->command = 0;
 				break;
@@ -1334,18 +1396,18 @@ static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 			ptr = ptr->v_next;
 		}
 
-		if ((lp->cmd_head != I596_NULL) && (dev->start))
+		if ((lp->cmd_head != I596_NULL))
 			ack_cmd |= CUC_START;
 		lp->scb.cmd = WSWAPcmd(virt_to_bus(&lp->cmd_head->status));
 	}
 	if ((status & 0x1000) || (status & 0x4000)) {
 		if ((status & 0x4000))
-			DEB(DEB_INTS,printk("%s: i596 interrupt received a frame.\n", dev->name));
+			DEB(DEB_INTS,printk(KERN_DEBUG "%s: i596 interrupt received a frame.\n", dev->name));
 		i596_rx(dev);
 		/* Only RX_START if stopped - RGH 07-07-96 */
 		if (status & 0x1000) {
-			if (dev->start) {
-				DEB(DEB_ERRORS,printk("%s: i596 interrupt receive unit inactive, status 0x%x\n", dev->name, status));
+			if (netif_running(dev)) {
+				DEB(DEB_ERRORS,printk(KERN_ERR "%s: i596 interrupt receive unit inactive, status 0x%x\n", dev->name, status));
 				ack_cmd |= RX_START;
 				lp->stats.rx_errors++;
 				lp->stats.rx_fifo_errors++;
@@ -1379,32 +1441,31 @@ static void i596_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 #endif
 	CA(dev);
 
-	DEB(DEB_INTS,printk("%s: exiting interrupt.\n", dev->name));
+	DEB(DEB_INTS,printk(KERN_DEBUG "%s: exiting interrupt.\n", dev->name));
 
-	dev->interrupt = 0;
-	return;
+	spin_unlock (&lp->lock);
+	return IRQ_RETVAL(handled);
 }
 
-static int i596_close(struct device *dev)
+static int i596_close(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 	unsigned long flags;
 
-	dev->start = 0;
-	dev->tbusy = 1;
+	netif_stop_queue(dev);
 
-	DEB(DEB_INIT,printk("%s: Shutting down ethercard, status was %4.4x.\n",
+	DEB(DEB_INIT,printk(KERN_DEBUG "%s: Shutting down ethercard, status was %4.4x.\n",
 		       dev->name, lp->scb.status));
 
-	save_flags(flags);
-	cli();
+	spin_lock_irqsave(&lp->lock, flags);
 
 	wait_cmd(dev,lp,100,"close1 timed out");
 	lp->scb.command = CUC_ABORT | RX_ABORT;
 	CA(dev);
 
 	wait_cmd(dev,lp,100,"close2 timed out");
-	restore_flags(flags);
+
+	spin_unlock_irqrestore(&lp->lock, flags);
 	DEB(DEB_STRUCT,i596_display_data(dev));
 	i596_cleanup_cmd(dev,lp);
 
@@ -1428,15 +1489,14 @@ static int i596_close(struct device *dev)
 
 	free_irq(dev->irq, dev);
 	remove_rx_bufs(dev);
-	MOD_DEC_USE_COUNT;
 
 	return 0;
 }
 
 static struct net_device_stats *
- i596_get_stats(struct device *dev)
+ i596_get_stats(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 
 	return &lp->stats;
 }
@@ -1445,12 +1505,18 @@ static struct net_device_stats *
  *    Set or clear the multicast filter for this adaptor.
  */
 
-static void set_multicast_list(struct device *dev)
+static void set_multicast_list(struct net_device *dev)
 {
-	struct i596_private *lp = (struct i596_private *) dev->priv;
+	struct i596_private *lp = dev->priv;
 	int config = 0, cnt;
 
-	DEB(DEB_MULTI,printk("%s: set multicast list, %d entries, promisc %s, allmulti %s\n", dev->name, dev->mc_count, dev->flags & IFF_PROMISC ? "ON" : "OFF", dev->flags & IFF_ALLMULTI ? "ON" : "OFF"));
+	DEB(DEB_MULTI,printk(KERN_DEBUG "%s: set multicast list, %d entries, promisc %s, allmulti %s\n",
+		dev->name, dev->mc_count,
+		dev->flags & IFF_PROMISC  ? "ON" : "OFF",
+		dev->flags & IFF_ALLMULTI ? "ON" : "OFF"));
+
+	if (wait_cfg(dev, &lp->cf_cmd.cmd, 1000, "config change request timed out"))
+		return;
 
 	if ((dev->flags & IFF_PROMISC) && !(lp->cf_cmd.i596_config[8] & 0x01)) {
 		lp->cf_cmd.i596_config[8] |= 0x01;
@@ -1469,20 +1535,15 @@ static void set_multicast_list(struct device *dev)
 		config = 1;
 	}
 	if (config) {
-		if (lp->cf_cmd.cmd.command)
-			printk("%s: config change request already queued\n",
-			       dev->name);
-		else {
-			lp->cf_cmd.cmd.command = CmdConfigure;
-			i596_add_cmd(dev, &lp->cf_cmd.cmd);
-		}
+		lp->cf_cmd.cmd.command = CmdConfigure;
+		i596_add_cmd(dev, &lp->cf_cmd.cmd);
 	}
 
 	cnt = dev->mc_count;
 	if (cnt > MAX_MC_CNT)
 	{
 		cnt = MAX_MC_CNT;
-		printk("%s: Only %d multicast addresses supported",
+		printk(KERN_ERR "%s: Only %d multicast addresses supported",
 			dev->name, cnt);
 	}
 	
@@ -1491,6 +1552,8 @@ static void set_multicast_list(struct device *dev)
 		unsigned char *cp;
 		struct mc_cmd *cmd;
 
+		if (wait_cfg(dev, &lp->mc_cmd.cmd, 1000, "multicast list change request timed out"))
+			return;
 		cmd = &lp->mc_cmd;
 		cmd->cmd.command = CmdMulticastList;
 		cmd->mc_cnt = dev->mc_count * 6;
@@ -1498,69 +1561,52 @@ static void set_multicast_list(struct device *dev)
 		for (dmi = dev->mc_list; cnt && dmi != NULL; dmi = dmi->next, cnt--, cp += 6) {
 			memcpy(cp, dmi->dmi_addr, 6);
 			if (i596_debug > 1)
-				DEB(DEB_MULTI,printk("%s: Adding address %02x:%02x:%02x:%02x:%02x:%02x\n",
+				DEB(DEB_MULTI,printk(KERN_INFO "%s: Adding address %02x:%02x:%02x:%02x:%02x:%02x\n",
 						dev->name, cp[0],cp[1],cp[2],cp[3],cp[4],cp[5]));
 		}
 		i596_add_cmd(dev, &cmd->cmd);
 	}
 }
 
-#ifdef HAVE_DEVLIST
-static unsigned int i596_portlist[] __initdata =
-{0x300, 0};
-struct netdev_entry i596_drv =
-{"i82596", i82596_probe, I596_TOTAL_SIZE, i596_portlist};
-#endif
-
 #ifdef MODULE
-static char devicename[9] =
-{0,};
-static struct device dev_82596 =
-{
-	devicename,	/* device name inserted by drivers/net/net_init.c */
-	0, 0, 0, 0,
-	0, 0,		/* base, irq */
-	0, 0, 0, NULL, i82596_probe};
+static struct net_device *dev_82596;
 
 #ifdef ENABLE_APRICOT
-static int io = 0x300;
-static int irq = 10;
 MODULE_PARM(irq, "i");
+MODULE_PARM_DESC(irq, "Apricot IRQ number");
 #endif
 
 MODULE_PARM(debug, "i");
+MODULE_PARM_DESC(debug, "i82596 debug mask");
 static int debug = -1;
 
 int init_module(void)
 {
-#ifdef ENABLE_APRICOT
-	dev_82596.base_addr = io;
-	dev_82596.irq = irq;
-#endif
 	if (debug >= 0)
 		i596_debug = debug;
-	if (register_netdev(&dev_82596) != 0)
-		return -EIO;
+	dev_82596 = i82596_probe(-1);
+	if (IS_ERR(dev_82596))
+		return PTR_ERR(dev_82596);
 	return 0;
 }
 
 void cleanup_module(void)
 {
-	unregister_netdev(&dev_82596);
+	unregister_netdev(dev_82596);
 #ifdef __mc68000__
 	/* XXX This assumes default cache mode to be IOMAP_FULL_CACHING,
 	 * XXX which may be invalid (CONFIG_060_WRITETHROUGH)
 	 */
 
-	kernel_set_cachemode((void *)(dev_82596.mem_start), 4096,
+	kernel_set_cachemode((void *)(dev_82596->mem_start), 4096,
 			IOMAP_FULL_CACHING);
 #endif
-	free_page ((u32)(dev_82596.mem_start));
-	dev_82596.priv = NULL;
+	free_page ((u32)(dev_82596->mem_start));
 #ifdef ENABLE_APRICOT
 	/* If we don't do this, we can't re-insmod it later. */
-	release_region(dev_82596.base_addr, I596_TOTAL_SIZE);
+	release_region(dev_82596->base_addr, I596_TOTAL_SIZE);
 #endif
+	free_netdev(dev_82596);
 }
 
 #endif				/* MODULE */
