@@ -5,15 +5,16 @@
  *
  */
 
-#include <linux/blkdev.h>
-#include <linux/init.h>
 #include <linux/module.h>
 #include <linux/slab.h>
-#include <linux/errno.h>
 #include <linux/fs.h>
+#include <linux/errno.h>
 #include <linux/stat.h>
-#include <linux/string.h>
 #include <linux/nls.h>
+#include <linux/buffer_head.h>
+#include <linux/vfs.h>
+#include <linux/parser.h>
+#include <linux/namei.h>
 
 #include "befs.h"
 #include "btree.h"
@@ -23,7 +24,6 @@
 #include "io.h"
 #include "endian.h"
 
-EXPORT_NO_SYMBOLS;
 MODULE_DESCRIPTION("BeOS File System (BeFS) driver");
 MODULE_AUTHOR("Will Dyson");
 MODULE_LICENSE("GPL");
@@ -32,74 +32,63 @@ MODULE_LICENSE("GPL");
 #define VFS_BLOCK_SIZE 512
 
 static int befs_readdir(struct file *, void *, filldir_t);
-static int befs_get_block(struct inode *, long, struct buffer_head *, int);
+static int befs_get_block(struct inode *, sector_t, struct buffer_head *, int);
 static int befs_readpage(struct file *file, struct page *page);
-static int befs_bmap(struct address_space *mapping, long block);
-static struct dentry *befs_lookup(struct inode *, struct dentry *);
+static sector_t befs_bmap(struct address_space *mapping, sector_t block);
+static struct dentry *befs_lookup(struct inode *, struct dentry *, struct nameidata *);
 static void befs_read_inode(struct inode *ino);
-static void befs_clear_inode(struct inode *ino);
+static struct inode *befs_alloc_inode(struct super_block *sb);
+static void befs_destroy_inode(struct inode *inode);
 static int befs_init_inodecache(void);
 static void befs_destroy_inodecache(void);
-
-static int befs_readlink(struct dentry *, char *, int);
-static int befs_follow_link(struct dentry *, struct nameidata *nd);
-
+static int befs_follow_link(struct dentry *, struct nameidata *);
+static void befs_put_link(struct dentry *, struct nameidata *);
 static int befs_utf2nls(struct super_block *sb, const char *in, int in_len,
 			char **out, int *out_len);
 static int befs_nls2utf(struct super_block *sb, const char *in, int in_len,
 			char **out, int *out_len);
-
 static void befs_put_super(struct super_block *);
-static struct super_block *befs_read_super(struct super_block *, void *, int);
 static int befs_remount(struct super_block *, int *, char *);
-static int befs_statfs(struct super_block *, struct statfs *);
+static int befs_statfs(struct super_block *, struct kstatfs *);
 static int parse_options(char *, befs_mount_options *);
 
-static ssize_t befs_listxattr(struct dentry *dentry, char *buffer, size_t size);
-static ssize_t befs_getxattr(struct dentry *dentry, const char *name,
-			     void *buffer, size_t size);
-static int befs_setxattr(struct dentry *dentry, const char *name, void *value,
-			 size_t size, int flags);
-static int befs_removexattr(struct dentry *dentry, const char *name);
+static const struct super_operations befs_sops = {
+	.read_inode	= befs_read_inode,	/* initialize & read inode */
+	.alloc_inode	= befs_alloc_inode,	/* allocate a new inode */
+	.destroy_inode	= befs_destroy_inode, /* deallocate an inode */
+	.put_super	= befs_put_super,	/* uninit super */
+	.statfs		= befs_statfs,	/* statfs */
+	.remount_fs	= befs_remount,
+};
 
 /* slab cache for befs_inode_info objects */
 static kmem_cache_t *befs_inode_cachep;
 
-static const struct super_operations befs_sops = {
-	read_inode:befs_read_inode,	/* initialize & read inode */
-	clear_inode:befs_clear_inode,	/* uninit inode */
-	put_super:befs_put_super,	/* uninit super */
-	statfs:befs_statfs,	/* statfs */
-	remount_fs:befs_remount,
-};
-
 struct file_operations befs_dir_operations = {
-	read:generic_read_dir,
-	readdir:befs_readdir,
+	.read		= generic_read_dir,
+	.readdir	= befs_readdir,
 };
 
 struct inode_operations befs_dir_inode_operations = {
-	lookup:befs_lookup,
+	.lookup		= befs_lookup,
 };
 
 struct file_operations befs_file_operations = {
-	llseek:default_llseek,
-	read:generic_file_read,
-	mmap:generic_file_mmap,
-};
-
-struct inode_operations befs_file_inode_operations = {
+	.llseek		= default_llseek,
+	.read		= generic_file_read,
+	.mmap		= generic_file_readonly_mmap,
 };
 
 struct address_space_operations befs_aops = {
-	readpage:befs_readpage,
-	sync_page:block_sync_page,
-	bmap:befs_bmap,
+	.readpage	= befs_readpage,
+	.sync_page	= block_sync_page,
+	.bmap		= befs_bmap,
 };
 
 static struct inode_operations befs_symlink_inode_operations = {
-	readlink:befs_readlink,
-	follow_link:befs_follow_link,
+	.readlink	= generic_readlink,
+	.follow_link	= befs_follow_link,
+	.put_link	= befs_put_link,
 };
 
 /* 
@@ -115,8 +104,8 @@ befs_readpage(struct file *file, struct page *page)
 	return block_read_full_page(page, befs_get_block);
 }
 
-static int
-befs_bmap(struct address_space *mapping, long block)
+static sector_t
+befs_bmap(struct address_space *mapping, sector_t block)
 {
 	return generic_block_bmap(mapping, block, befs_get_block);
 }
@@ -133,7 +122,7 @@ befs_bmap(struct address_space *mapping, long block)
  */
 
 static int
-befs_get_block(struct inode *inode, long block,
+befs_get_block(struct inode *inode, sector_t block,
 	       struct buffer_head *bh_result, int create)
 {
 	struct super_block *sb = inode->i_sb;
@@ -168,9 +157,7 @@ befs_get_block(struct inode *inode, long block,
 
 	disk_off = (ulong) iaddr2blockno(sb, &run);
 
-	bh_result->b_dev = inode->i_dev;
-	bh_result->b_blocknr = disk_off;
-	bh_result->b_state |= (1UL << BH_Mapped);
+	map_bh(bh_result, inode->i_sb, disk_off);
 
 	befs_debug(sb, "<--- befs_get_block() for inode %lu, block %ld, "
 		   "disk address %lu", inode->i_ino, block, disk_off);
@@ -179,7 +166,7 @@ befs_get_block(struct inode *inode, long block,
 }
 
 static struct dentry *
-befs_lookup(struct inode *dir, struct dentry *dentry)
+befs_lookup(struct inode *dir, struct dentry *dentry, struct nameidata *nd)
 {
 	struct inode *inode = NULL;
 	struct super_block *sb = dir->i_sb;
@@ -292,16 +279,31 @@ befs_readdir(struct file *filp, void *dirent, filldir_t filldir)
 	return 0;
 }
 
-static void
-befs_clear_inode(struct inode *inode)
+static struct inode *
+befs_alloc_inode(struct super_block *sb)
 {
-	befs_inode_info *b_ino = BEFS_I(inode);
-	inode->u.generic_ip = NULL;
+        struct befs_inode_info *bi;
+        bi = (struct befs_inode_info *)kmem_cache_alloc(befs_inode_cachep,
+							SLAB_KERNEL);
+        if (!bi)
+                return NULL;
+        return &bi->vfs_inode;
+}
 
-	if (b_ino) {
-		kmem_cache_free(befs_inode_cachep, b_ino);
-	}
-	return;
+static void
+befs_destroy_inode(struct inode *inode)
+{
+        kmem_cache_free(befs_inode_cachep, BEFS_I(inode));
+}
+
+static void init_once(void * foo, kmem_cache_t * cachep, unsigned long flags)
+{
+        struct befs_inode_info *bi = (struct befs_inode_info *) foo;
+	
+	        if ((flags & (SLAB_CTOR_VERIFY|SLAB_CTOR_CONSTRUCTOR)) ==
+		            SLAB_CTOR_CONSTRUCTOR) {
+			inode_init_once(&bi->vfs_inode);
+		}
 }
 
 static void
@@ -316,12 +318,6 @@ befs_read_inode(struct inode *inode)
 
 	befs_debug(sb, "---> befs_read_inode() " "inode = %lu", inode->i_ino);
 
-	inode->u.generic_ip = kmem_cache_alloc(befs_inode_cachep, GFP_NOFS);
-	if (inode->u.generic_ip == NULL) {
-		befs_error(sb, "Unable to allocate memory for private "
-			   "portion of inode %lu.", inode->i_ino);
-		goto unaquire_none;
-	}
 	befs_ino = BEFS_I(inode);
 
 	/* convert from vfs's inode number to befs's inode number */
@@ -331,11 +327,11 @@ befs_read_inode(struct inode *inode)
 		   befs_ino->i_inode_num.allocation_group,
 		   befs_ino->i_inode_num.start, befs_ino->i_inode_num.len);
 
-	bh = befs_bread_iaddr(sb, befs_ino->i_inode_num);
+	bh = befs_bread(sb, inode->i_ino);
 	if (!bh) {
 		befs_error(sb, "unable to read inode block - "
 			   "inode = %lu", inode->i_ino);
-		goto unaquire_ino_info;
+		goto unaquire_none;
 	}
 
 	raw_inode = (befs_inode *) bh->b_data;
@@ -352,7 +348,7 @@ befs_read_inode(struct inode *inode)
 	/*
 	 * set uid and gid.  But since current BeOS is single user OS, so
 	 * you can change by "uid" or "gid" options.
-	 */
+	 */   
 
 	inode->i_uid = befs_sb->mount_opts.use_uid ?
 	    befs_sb->mount_opts.uid : (uid_t) fs32_to_cpu(sb, raw_inode->uid);
@@ -365,13 +361,16 @@ befs_read_inode(struct inode *inode)
 	 * BEFS's time is 64 bits, but current VFS is 32 bits...
 	 * BEFS don't have access time. Nor inode change time. VFS
 	 * doesn't have creation time.
+	 * Also, the lower 16 bits of the last_modified_time and 
+	 * create_time are just a counter to help ensure uniqueness
+	 * for indexing purposes. (PFD, page 54)
 	 */
 
-	inode->i_mtime =
-	    (time_t) (fs64_to_cpu(sb, raw_inode->last_modified_time) >> 16);
+	inode->i_mtime.tv_sec =
+	    fs64_to_cpu(sb, raw_inode->last_modified_time) >> 16;
+	inode->i_mtime.tv_nsec = 0;   /* lower 16 bits are not a time */	
 	inode->i_ctime = inode->i_mtime;
 	inode->i_atime = inode->i_mtime;
-	inode->i_blkbits = befs_sb->block_shift;
 	inode->i_blksize = befs_sb->block_size;
 
 	befs_ino->i_inode_num = fsrun_to_cpu(sb, raw_inode->inode_num);
@@ -379,7 +378,7 @@ befs_read_inode(struct inode *inode)
 	befs_ino->i_attribute = fsrun_to_cpu(sb, raw_inode->attributes);
 	befs_ino->i_flags = fs32_to_cpu(sb, raw_inode->flags);
 
-	if (S_ISLNK(inode->i_mode) && !(inode->i_flags & BEFS_LONG_SYMLINK)) {
+	if (S_ISLNK(inode->i_mode) && !(befs_ino->i_flags & BEFS_LONG_SYMLINK)){
 		inode->i_size = 0;
 		inode->i_blocks = befs_sb->block_size / VFS_BLOCK_SIZE;
 		strncpy(befs_ino->i_data.symlink, raw_inode->data.symlink,
@@ -400,7 +399,6 @@ befs_read_inode(struct inode *inode)
 
 	if (S_ISREG(inode->i_mode)) {
 		inode->i_fop = &befs_file_operations;
-		inode->i_op = &befs_file_inode_operations;
 	} else if (S_ISDIR(inode->i_mode)) {
 		inode->i_op = &befs_dir_inode_operations;
 		inode->i_fop = &befs_dir_operations;
@@ -420,12 +418,8 @@ befs_read_inode(struct inode *inode)
       unaquire_bh:
 	brelse(bh);
 
-      unaquire_ino_info:
-	kmem_cache_free(befs_inode_cachep, inode->u.generic_ip);
-
       unaquire_none:
 	make_bad_inode(inode);
-	inode->u.generic_ip = NULL;
 	befs_debug(sb, "<--- befs_read_inode() - Bad inode");
 	return;
 }
@@ -439,8 +433,8 @@ befs_init_inodecache(void)
 {
 	befs_inode_cachep = kmem_cache_create("befs_inode_cache",
 					      sizeof (struct befs_inode_info),
-					      0, SLAB_HWCACHE_ALIGN,
-					      NULL, NULL);
+					      0, SLAB_HWCACHE_ALIGN|SLAB_RECLAIM_ACCOUNT,
+					      init_once, NULL);
 	if (befs_inode_cachep == NULL) {
 		printk(KERN_ERR "befs_init_inodecache: "
 		       "Couldn't initalize inode slabcache\n");
@@ -470,75 +464,45 @@ befs_destroy_inodecache(void)
 static int
 befs_follow_link(struct dentry *dentry, struct nameidata *nd)
 {
-	struct super_block *sb = dentry->d_sb;
 	befs_inode_info *befs_ino = BEFS_I(dentry->d_inode);
 	char *link;
-	int res;
 
 	if (befs_ino->i_flags & BEFS_LONG_SYMLINK) {
+		struct super_block *sb = dentry->d_sb;
 		befs_data_stream *data = &befs_ino->i_data.ds;
-		befs_off_t linklen = data->size;
+		befs_off_t len = data->size;
 
 		befs_debug(sb, "Follow long symlink");
 
-		link = kmalloc(linklen, GFP_NOFS);
-		if (link == NULL)
-			return -ENOMEM;
-
-		if (befs_read_lsymlink(sb, data, link, linklen) != linklen) {
+		link = kmalloc(len, GFP_NOFS);
+		if (!link) {
+			link = ERR_PTR(-ENOMEM);
+		} else if (befs_read_lsymlink(sb, data, link, len) != len) {
 			kfree(link);
 			befs_error(sb, "Failed to read entire long symlink");
-			return -EIO;
+			link = ERR_PTR(-EIO);
 		}
-
-		res = vfs_follow_link(nd, link);
-
-		kfree(link);
 	} else {
 		link = befs_ino->i_data.symlink;
-		res = vfs_follow_link(nd, link);
 	}
 
-	return res;
+	nd_set_link(nd, link);
+	return 0;
 }
 
-static int
-befs_readlink(struct dentry *dentry, char *buffer, int buflen)
+static void befs_put_link(struct dentry *dentry, struct nameidata *nd)
 {
-	struct super_block *sb = dentry->d_sb;
 	befs_inode_info *befs_ino = BEFS_I(dentry->d_inode);
-	char *link;
-	int res;
-
 	if (befs_ino->i_flags & BEFS_LONG_SYMLINK) {
-		befs_data_stream *data = &befs_ino->i_data.ds;
-		befs_off_t linklen = data->size;
-
-		befs_debug(sb, "Read long symlink");
-
-		link = kmalloc(linklen, GFP_NOFS);
-		if (link == NULL)
-			return -ENOMEM;
-
-		if (befs_read_lsymlink(sb, data, link, linklen) != linklen) {
-			kfree(link);
-			befs_error(sb, "Failed to read entire long symlink");
-			return -EIO;
-		}
-
-		res = vfs_readlink(dentry, buffer, buflen, link);
-
-		kfree(link);
-	} else {
-		link = befs_ino->i_data.symlink;
-		res = vfs_readlink(dentry, buffer, buflen, link);
+		char *p = nd_get_link(nd);
+		if (!IS_ERR(p))
+			kfree(p);
 	}
-
-	return res;
 }
 
 /*
  * UTF-8 to NLS charset  convert routine
+ * 
  *
  * Changed 8/10/01 by Will Dyson. Now use uni2char() / char2uni() rather than
  * the nls tables directly
@@ -553,7 +517,7 @@ befs_utf2nls(struct super_block *sb, const char *in,
 	wchar_t uni;
 	int unilen, utflen;
 	char *result;
-	int maxlen = in_len;	/* The utf8->nls conversion cant make more chars */
+	int maxlen = in_len; /* The utf8->nls conversion can't make more chars */
 
 	befs_debug(sb, "---> utf2nls()");
 
@@ -578,20 +542,20 @@ befs_utf2nls(struct super_block *sb, const char *in,
 		}
 
 		/* convert from Unicode to nls */
-		unilen = nls->uni2char(uni, &result[o], 1);
+		unilen = nls->uni2char(uni, &result[o], in_len - o);
 		if (unilen < 0) {
 			goto conv_err;
 		}
 	}
 	result[o] = '\0';
+	*out_len = o;
 
 	befs_debug(sb, "<--- utf2nls()");
 
 	return o;
-	*out_len = o;
 
       conv_err:
-	befs_error(sb, "Name using charecter set %s contains a charecter that "
+	befs_error(sb, "Name using character set %s contains a character that "
 		   "cannot be converted to unicode.", nls->charset);
 	befs_debug(sb, "<--- utf2nls()");
 	kfree(result);
@@ -614,8 +578,8 @@ befs_utf2nls(struct super_block *sb, const char *in,
  * 
  * On return, *@destlen is the length of @dest in bytes.
  *
- * On success, the return value is the number of utf8 charecters written to
- * the ouput buffer @dest.
+ * On success, the return value is the number of utf8 characters written to
+ * the output buffer @dest.
  *  
  * On Failure, a negative number coresponding to the error code is returned.
  */
@@ -675,43 +639,28 @@ befs_nls2utf(struct super_block *sb, const char *in,
 	return -EILSEQ;
 }
 
-/****Xattr****/
+/**
+ * Use the
+ *
+ */
+enum {
+	Opt_uid, Opt_gid, Opt_charset, Opt_debug, Opt_err,
+};
 
-static ssize_t
-befs_listxattr(struct dentry *dentry, char *buffer, size_t size)
-{
-	printk(KERN_ERR "befs_listxattr called\n");
-	return 0;
-}
-
-static ssize_t
-befs_getxattr(struct dentry *dentry, const char *name,
-	      void *buffer, size_t size)
-{
-	return 0;
-}
-
-static int
-befs_setxattr(struct dentry *dentry, const char *name,
-	      void *value, size_t size, int flags)
-{
-	return 0;
-}
-
-static int
-befs_removexattr(struct dentry *dentry, const char *name)
-{
-	return 0;
-}
-
-/****Superblock****/
+static match_table_t befs_tokens = {
+	{Opt_uid, "uid=%d"},
+	{Opt_gid, "gid=%d"},
+	{Opt_charset, "iocharset=%s"},
+	{Opt_debug, "debug"},
+	{Opt_err, NULL}
+};
 
 static int
 parse_options(char *options, befs_mount_options * opts)
 {
-	char *this_char;
-	char *value;
-	int ret = 1;
+	char *p;
+	substring_t args[MAX_OPT_ARGS];
+	int option;
 
 	/* Initialize options */
 	opts->uid = 0;
@@ -722,65 +671,56 @@ parse_options(char *options, befs_mount_options * opts)
 	opts->debug = 0;
 
 	if (!options)
-		return ret;
+		return 1;
 
-	for (this_char = strtok(options, ","); this_char != NULL;
-	     this_char = strtok(NULL, ",")) {
+	while ((p = strsep(&options, ",")) != NULL) {
+		int token;
+		if (!*p)
+			continue;
 
-		if ((value = strchr(this_char, '=')) != NULL)
-			*value++ = 0;
-
-		if (!strcmp(this_char, "uid")) {
-			if (!value || !*value) {
-				ret = 0;
-			} else {
-				opts->uid = simple_strtoul(value, &value, 0);
-				opts->use_uid = 1;
-				if (*value) {
-					printk(KERN_ERR "BEFS: Invalid uid "
-					       "option: %s\n", value);
-					ret = 0;
-				}
+		token = match_token(p, befs_tokens, args);
+		switch (token) {
+		case Opt_uid:
+			if (match_int(&args[0], &option))
+				return 0;
+			if (option < 0) {
+				printk(KERN_ERR "BeFS: Invalid uid %d, "
+						"using default\n", option);
+				break;
 			}
-		} else if (!strcmp(this_char, "gid")) {
-			if (!value || !*value)
-				ret = 0;
-			else {
-				opts->gid = simple_strtoul(value, &value, 0);
-				opts->use_gid = 1;
-				if (*value) {
-					printk(KERN_ERR
-					       "BEFS: Invalid gid option: "
-					       "%s\n", value);
-					ret = 0;
-				}
+			opts->uid = option;
+			opts->use_uid = 1;
+			break;
+		case Opt_gid:
+			if (match_int(&args[0], &option))
+				return 0;
+			if (option < 0) {
+				printk(KERN_ERR "BeFS: Invalid gid %d, "
+						"using default\n", option);
+				break;
 			}
-		} else if (!strcmp(this_char, "iocharset") && value) {
-			char *p = value;
-			int len;
-
-			while (*value && *value != ',')
-				value++;
-			len = value - p;
-			if (len) {
-				char *buffer = kmalloc(len + 1, GFP_NOFS);
-				if (buffer) {
-					opts->iocharset = buffer;
-					memcpy(buffer, p, len);
-					buffer[len] = 0;
-
-				} else {
-					printk(KERN_ERR "BEFS: "
-					       "cannot allocate memory\n");
-					ret = 0;
-				}
+			opts->gid = option;
+			opts->use_gid = 1;
+			break;
+		case Opt_charset:
+			kfree(opts->iocharset);
+			opts->iocharset = match_strdup(&args[0]);
+			if (!opts->iocharset) {
+				printk(KERN_ERR "BeFS: allocation failure for "
+						"iocharset string\n");
+				return 0;
 			}
-		} else if (!strcmp(this_char, "debug")) {
+			break;
+		case Opt_debug:
 			opts->debug = 1;
+			break;
+		default:
+			printk(KERN_ERR "BeFS: Unrecognized mount option \"%s\" "
+					"or missing value\n", p);
+			return 0;
 		}
 	}
-
-	return ret;
+	return 1;
 }
 
 /* This function has the responsibiltiy of getting the
@@ -801,9 +741,9 @@ befs_put_super(struct super_block *sb)
 		BEFS_SB(sb)->nls = NULL;
 	}
 
-	if (sb->u.generic_sbp) {
-		kfree(sb->u.generic_sbp);
-		sb->u.generic_sbp = NULL;
+	if (sb->s_fs_info) {
+		kfree(sb->s_fs_info);
+		sb->s_fs_info = NULL;
 	}
 	return;
 }
@@ -814,33 +754,33 @@ befs_put_super(struct super_block *sb)
  * Make the root directory
  * Load a set of NLS translations if needed.
  */
-static struct super_block *
-befs_read_super(struct super_block *sb, void *data, int silent)
+static int
+befs_fill_super(struct super_block *sb, void *data, int silent)
 {
 	struct buffer_head *bh;
 	befs_sb_info *befs_sb;
 	befs_super_block *disk_sb;
-	int blocksize;
+	struct inode *root;
 
 	const unsigned long sb_block = 0;
 	const off_t x86_sb_off = 512;
 
-	sb->u.generic_sbp = kmalloc(sizeof (struct befs_sb_info), GFP_NOFS);
-	if (sb->u.generic_sbp == NULL) {
+	sb->s_fs_info = kmalloc(sizeof (*befs_sb), GFP_KERNEL);
+	if (sb->s_fs_info == NULL) {
 		printk(KERN_ERR
 		       "BeFS(%s): Unable to allocate memory for private "
-		       "portion of superblock. Bailing.\n",
-		       kdevname(sb->s_dev));
+		       "portion of superblock. Bailing.\n", sb->s_id);
 		goto unaquire_none;
 	}
 	befs_sb = BEFS_SB(sb);
+	memset(befs_sb, 0, sizeof(befs_sb_info));
 
 	if (!parse_options((char *) data, &befs_sb->mount_opts)) {
 		befs_error(sb, "cannot parse mount options");
 		goto unaquire_priv_sbp;
 	}
 
-	befs_debug(sb, "---> befs_read_super()");
+	befs_debug(sb, "---> befs_fill_super()");
 
 #ifndef CONFIG_BEFS_RW
 	if (!(sb->s_flags & MS_RDONLY)) {
@@ -858,11 +798,10 @@ befs_read_super(struct super_block *sb, void *data, int silent)
 	 * the hardsect size for the device. But we also need to read at 
 	 * least 1k to get the second 512 bytes of the volume.
 	 * -WD 10-26-01
-	 */
-	blocksize = max_t(int, get_hardsect_size(sb->s_dev), 1024);
-	set_blocksize(sb->s_dev, blocksize);
+	 */ 
+	sb_min_blocksize(sb, 1024);
 
-	if (!(bh = bread(sb->s_dev, sb_block, blocksize))) {
+	if (!(bh = sb_bread(sb, sb_block))) {
 		befs_error(sb, "unable to read superblock");
 		goto unaquire_priv_sbp;
 	}
@@ -888,17 +827,25 @@ befs_read_super(struct super_block *sb, void *data, int silent)
 	if (befs_check_sb(sb) != BEFS_OK)
 		goto unaquire_priv_sbp;
 
+	if( befs_sb->num_blocks > ~((sector_t)0) ) {
+		befs_error(sb, "blocks count: %Lu "
+			"is larger than the host can use",
+			befs_sb->num_blocks);
+		goto unaquire_priv_sbp;
+	}
+
 	/*
 	 * set up enough so that it can read an inode
 	 * Fill in kernel superblock fields from private sb
 	 */
 	sb->s_magic = BEFS_SUPER_MAGIC;
-	sb->s_blocksize = (ulong) befs_sb->block_size;
-	sb->s_blocksize_bits = (unsigned char) befs_sb->block_shift;
+	/* Set real blocksize of fs */
+	sb_set_blocksize(sb, (ulong) befs_sb->block_size);
 	sb->s_op = (struct super_operations *) &befs_sops;
-	sb->s_root =
-	    d_alloc_root(iget(sb, iaddr2blockno(sb, &(befs_sb->root_dir))));
+	root = iget(sb, iaddr2blockno(sb, &(befs_sb->root_dir)));
+	sb->s_root = d_alloc_root(root);
 	if (!sb->s_root) {
+		iput(root);
 		befs_error(sb, "get root inode failed");
 		goto unaquire_priv_sbp;
 	}
@@ -916,21 +863,17 @@ befs_read_super(struct super_block *sb, void *data, int silent)
 		}
 	}
 
-	/* Set real blocksize of fs */
-	set_blocksize(sb->s_dev, (int) befs_sb->block_size);
-
-	return sb;
+	return 0;
 /*****************/
       unaquire_bh:
 	brelse(bh);
 
       unaquire_priv_sbp:
-	kfree(sb->u.generic_sbp);
+	kfree(sb->s_fs_info);
 
       unaquire_none:
-	sb->s_dev = 0;
-	sb->u.generic_sbp = NULL;
-	return NULL;
+	sb->s_fs_info = NULL;
+	return -EINVAL;
 }
 
 static int
@@ -942,7 +885,7 @@ befs_remount(struct super_block *sb, int *flags, char *data)
 }
 
 static int
-befs_statfs(struct super_block *sb, struct statfs *buf)
+befs_statfs(struct super_block *sb, struct kstatfs *buf)
 {
 
 	befs_debug(sb, "---> befs_statfs()");
@@ -961,14 +904,20 @@ befs_statfs(struct super_block *sb, struct statfs *buf)
 	return 0;
 }
 
-/*
-	Makes a variable of type file_system_type, 
-	named befs_fs_tipe, identified by the "befs" string,
-	and containing a reference to the befs_read_super function
-	
-	Macro declared in <linux/fs.h>
-*/
-static DECLARE_FSTYPE_DEV(befs_fs_type, "befs", befs_read_super);
+static struct super_block *
+befs_get_sb(struct file_system_type *fs_type, int flags, const char *dev_name,
+	    void *data)
+{
+	return get_sb_bdev(fs_type, flags, dev_name, data, befs_fill_super);
+}
+
+static struct file_system_type befs_fs_type = {
+	.owner		= THIS_MODULE,
+	.name		= "befs",
+	.get_sb		= befs_get_sb,
+	.kill_sb	= kill_block_super,
+	.fs_flags	= FS_REQUIRES_DEV,	
+};
 
 static int __init
 init_befs_fs(void)
@@ -979,9 +928,19 @@ init_befs_fs(void)
 
 	err = befs_init_inodecache();
 	if (err)
-		return err;
+		goto unaquire_none;
 
-	return register_filesystem(&befs_fs_type);
+	err = register_filesystem(&befs_fs_type);
+	if (err)
+		goto unaquire_inodecache;
+
+	return 0;
+
+unaquire_inodecache:
+	befs_destroy_inodecache();
+
+unaquire_none:
+	return err;
 }
 
 static void __exit
@@ -998,4 +957,4 @@ ensures that they are called at init and cleanup,
 and eliminates warnings about unused functions.
 */
 module_init(init_befs_fs)
-    module_exit(exit_befs_fs)
+module_exit(exit_befs_fs)
