@@ -689,14 +689,9 @@ struct dentry *umsdos_solve_hlink (struct dentry *hlink)
 	struct dentry *dentry_dst;
 	char *path, *pt;
 	int len;
-	struct address_space *mapping;
+	struct address_space *mapping = hlink->d_inode->i_mapping;
 	struct page *page;
 
-	dentry_dst = ERR_PTR(-ENOENT);
-	if (!hlink->d_inode)
-		goto out;
-
-	mapping = hlink->d_inode->i_mapping;
 	page=read_cache_page(mapping,0,(filler_t *)mapping->a_ops->readpage,NULL);
 	dentry_dst=(struct dentry *)page;
 	if (IS_ERR(page))
